/*	  Copyright (C) 2001  Sony Computer Entertainment America
       	  
       	  This file is subject to the terms and conditions of the GNU General
       	  Public License Version 2. See the file "COPYING" in the main
       	  directory of this archive for more details.                               */

#ifndef ps2s_core_h
#define ps2s_core_h

#include <malloc.h>

#include "ps2s/types.h"
#include "ps2s/debug.h"

namespace Core {

   /********************************************
    * MemMappings
    */

   namespace MemMappings
   {
      static const tU32 Normal = 0x00000000,
//  	 Uncached = 0x20000000,
//  	 UncachedAccl = 0x30000000,

	 // these can't be used in user space, so temporarily
	 // disable them for linux..
	 Uncached = 0x00000000,
	 UncachedAccl = 0x00000000,

	 SP = 0x70000000,
	 VU0Code = 0x11000000,
	 VU0Data = 0x11004000,
	 VU1Code = 0x11008000,
	 VU1Data = 0x1100c000;
   }

   template <class ptrType>
   inline ptrType MakePtrNormal( ptrType ptr ) {
      // since we're in a virtual address space, the upper 4 bits
      // are valid address bits, so we can't mask them off...
      // return reinterpret_cast<ptrType>( (tU32)ptr & 0x0fffffff );
      return reinterpret_cast<ptrType>( (tU32)ptr & 0xffffffff );
   }

   template <class ptrType>
   inline ptrType MakePtrUncached( ptrType ptr ) {
      return reinterpret_cast<ptrType>( (tU32)MakePtrNormal(ptr) | MemMappings::Uncached );
   }

   template <class ptrType>
   inline ptrType MakePtrUncachedAccl( ptrType ptr ) {
      return reinterpret_cast<ptrType>( (tU32)MakePtrNormal(ptr) | MemMappings::UncachedAccl);
   }

   /********************************************
    * misc
    */

   // when "operator new" needs to return qword-aligned memory

   inline void *New16( size_t size ) {
      size = (size == 0) ? 1 : size;
      void *mem = memalign( 16, size );
      mErrorIf( mem == NULL, "Could not allocate memory, exiting." );
      return mem;
   }

   // cop0 counter

   inline void ZeroCount( void );
   inline tU32 GetCount( void );

   // fpu

   inline tU32 FToI4( float flp );
	
   // cop0 performance counter

   static	const	u_int	COP0_NUM_PERF_COUNTERS	= 2;
   static	const	u_int	COP0_NUM_PERF_EVENTS	= 17;

   void		SetupPerfCounters(u_int evt_0,u_int evt_1);	// set up,also zero + halt

   inline	void	HaltPerfCounters();				// stop both counters	
   inline	void	ZeroStartPerfCounters();			// clear and start both counters

   inline	u_int	ReadPerfCounter0();				// read counter 0
   inline	u_int	ReadPerfCounter1();				// read counter 1
		
   // names of the events for people

   extern		const	char   	*ev0_Name[COP0_NUM_PERF_EVENTS];
   extern		const	char   	*ev1_Name[COP0_NUM_PERF_EVENTS];

} 

/********************************************
 * Core inlines
 */

inline void
Core::ZeroCount( void ) {
   asm __volatile__ ("
      mtc0	$0, $9
      sync.p
   ");
}

inline tU32
Core::GetCount( void ) {
   tU32 ret;

   asm __volatile__ ("
      mfc0	%0, $9
      sync.p
   " : "=r" (ret) );

   return ret;
}

inline tU32
Core::FToI4( float flp ) {
   tU32 fip;

   __asm__ __volatile__ ("
      qmtc2	%1, vf01
      vftoi4	vf01, vf01
      qmfc2	%0, vf01
   " : "=r" (fip)
			 : "r" (flp) );

   return fip;
}

// performance counter inlines

inline	void	Core::HaltPerfCounters()	
{
   asm ("
			mtps	$0,0
			sync.p
		");
};
																						
inline	void	Core::ZeroStartPerfCounters()
{
   asm	__volatile__("
			mtps	$0,0			# halt performance counters
			sync.p				# 
			mtpc	$0,0 			# set perfcounter 0 to zero
			mtpc	$0,1 			# set perfcounter 1 to zero
			sync.p				# 
			lui		v0,0x8000	    	# master enable
			mtps	v0,0 			# truly - we rule
			"	
		     :	// no output
		     :	// no input
		     :	"v0");
};

inline	u_int	Core::ReadPerfCounter0()
{
   u_int	ret;
   asm	__volatile__("
		mfpc	%0,0
		sync.p
		"
		     : "=r" (ret) );	
   return	ret;	
}

inline	u_int	Core::ReadPerfCounter1()
{
   u_int	ret;
   asm	__volatile__("
	
		mfpc	%0,1
		sync.p
	
		" 	: "=r" (ret) );
   return	ret;	
}
	
#endif // ps2s_core_h
