/*	  Copyright (C) 2001  Sony Computer Entertainment America
       	  
       	  This file is subject to the terms and conditions of the GNU General
       	  Public License Version 2. See the file "COPYING" in the main
       	  directory of this archive for more details.                               */

#ifndef ps2s_drawenv_h
#define ps2s_drawenv_h

/********************************************
 * includes
 */

#include "ps2s/types.h"
#include "ps2s/core.h"
#include "ps2s/dmac.h"
#include "ps2s/gs.h"
#include "ps2s/packet.h"

namespace GS {

   /********************************************
    * constants
    */

   static const bool	kDontAddHalfPixel = false;
   static const bool	kAddHalfPixel = true;

   /********************************************
    * typedefs
    */

   // the "using"s after each namespace { ... } are to make the type
   // available from the GS namespace (ie, so you can say "GS::tAlphaBlendVal")

   namespace ABlend {
      typedef enum { 
	 kSourceRGB = 0,
	 kDestRGB = 1,
	 kSourceAlpha = 0,
	 kDestAlpha = 1,
	 kZero = 2,
	 kFix = 2
      } tAlphaBlendVal;
   }
   using ABlend::tAlphaBlendVal;

   namespace ATest {
      typedef enum {
	 kNever,
	 kAlways,
	 kLess,
	 kLEqual,
	 kEqual,
	 kGEqual,
	 kGreater,
	 kNotEqual
      } tAlphaTestPassMode;

      typedef enum {
	 kKeep,
	 kFrameBuffOnly,
	 kZBuffOnly,
	 kRGBOnly
      } tAlphaTestFailAction;
   }
   using ATest::tAlphaTestPassMode;
   using ATest::tAlphaTestFailAction;

   namespace ZTest {
      typedef enum {
	 kNever,
	 kAlways,
	 kGEqual,
	 kGreater
      } tZTestPassMode;
   }
   using ZTest::tZTestPassMode;

   namespace Dither {
     typedef enum {
       kMinusFour = 0x4, 
       kMinusThree = 0x5,
       kMinusTwo = 0x6, 
       kMinusOne = 0x7,
       kZero = 0,
       kOne = 1,
       kTwo = 2,
       kThree = 3
     } tDitherVal;
   }
   using Dither::tDitherVal;

   /********************************************
    * class definition
    */

   class CDrawEnv {
      public:
	 CDrawEnv( GS::tContext context );
	 CDrawEnv( const CDrawEnv& rhs );
	 CDrawEnv( GS::tContext context, tU32 fbW, tU32 fbH, tU32 fbWordAddr, tU32 zbufWordAddr = NULL );

	 void operator=( const CDrawEnv& otherDE );

	 void* operator new( size_t size ) {
	    return Core::New16(size);
	 }

	 void SetContext( GS::tContext context );

	 inline void EnableAlphaTest( void ) { gsrTest.ATE = 1; }
	 inline void DisableAlphaTest( void ) { gsrTest.ATE = 0; }
	 inline void SetAlphaTestPassMode( tAlphaTestPassMode mode ) { gsrTest.ATST = (tU64)mode; }
	 inline void SetAlphaTestFailAction( tAlphaTestFailAction action ) { gsrTest.AFAIL = (tU64)action; }
	 inline void SetAlphaRefVal( tU32 refVal ) { gsrTest.AREF = refVal; }

	 inline void EnableSelectiveAlphaBlend( void ) { gsrPABE.PABE = 1; }
	 inline void DisableSelectiveAlphaBlend( void ) { gsrPABE.PABE = 0; }

	 inline void SetAlphaBlendFunc( tAlphaBlendVal a, tAlphaBlendVal b, tAlphaBlendVal c, tAlphaBlendVal d, tU32 fix );

	 inline void EnableColorClamp( void ) { gsrColClamp.CLAMP = 1; }
	 inline void DisableColorClamp( void ) { gsrColClamp.CLAMP = 0; }

	 inline void EnableDestAlphaTest( void ) { gsrTest.DATE = 1; }
	 inline void DisableDestAlphaTest( void ) { gsrTest.DATE = 0; }
	 inline void SetDestAlpha0Pass( void ) { gsrTest.DATM = 0; }
	 inline void SetDestAlpha1Pass( void ) { gsrTest.DATM = 1; }

	 inline void SetDepthTestPassMode( tZTestPassMode passMode ) { gsrTest.ZTST = (tU64)passMode; eZTestPassMode = passMode; }
	 inline void SetDepthBufferAddr( tU32 wordAddress ) { gsrZBuf.ZBP = wordAddress/2048; }
	 inline void SetDepthBufferPSM( tU32 psm ) { gsrZBuf.PSM = (tU64)psm; }
	 inline void EnableDepthTest( void ) { gsrZBuf.ZMSK = 0; SetDepthTestPassMode(eZTestPassMode); }
	 inline void DisableDepthTest( void ) { gsrZBuf.ZMSK = 1; gsrTest.ZTST = (tU64)ZTest::kAlways; }
	 inline void SetDepthWriteEnabled( bool write ) { gsrZBuf.ZMSK = ! write; }

	 inline void SetFrameBufferAddr( tU32 wordAddress ) { mAssert( (wordAddress & 2047) == 0 ); gsrFrame.FBP = wordAddress/2048; }
	 inline void SetFrameBufferDrawMask( tU32 drawMask ) { gsrFrame.FBMSK = drawMask; }
	 void SetFrameBufferDim( tU32 pixelW, tU32 pixelH );
	 inline void SetFrameBufferPSM( tU32 psm ) { gsrFrame.PSM = psm; }
	 float GetInterlacedPixelOffset() const { return InterlacedOffset; }
	 void SetInterlacedPixelOffset( float offset ) { InterlacedOffset = offset; }
	 void CalculateClippedFBXYOffsets( bool addHalfPixel );
	 inline void SetXYOffsets( tU16 offsetX, tU16 offsetY ) { gsrXYOffset.OFX = offsetX << 4; gsrXYOffset.OFY = offsetY << 4; }
	 inline void SetXYOffsetsFix4( tU16 offsetX, tU16 offsetY ) { gsrXYOffset.OFX = offsetX; gsrXYOffset.OFY = offsetY; }

	 inline void SetFrameBufAlphaORMask( tU32 mask ) { gsrFBA.FBA = mask; }
	 inline void SetFogColor( tU32 r, tU32 g, tU32 b ) { gsrFogCol.FCR = r; gsrFogCol.FCG = g; gsrFogCol.FCB = b; }

	 inline void EnableDither( void ) {  gsrDTHE.DTHE = 1; }
	 inline void DisableDither( void ) { gsrDTHE.DTHE = 0; }
	 inline void SetDitherMatrix( tDitherVal dm00, tDitherVal dm01, tDitherVal dm02, tDitherVal dm03, 
				      tDitherVal dm10, tDitherVal dm11, tDitherVal dm12, tDitherVal dm13, 
				      tDitherVal dm20, tDitherVal dm21, tDitherVal dm22, tDitherVal dm23, 
				      tDitherVal dm30, tDitherVal dm31, tDitherVal dm32, tDitherVal dm33 );

	 inline void SetScissorArea( tU32 scX, tU32 scY, tU32 scWidth, tU32 scHeight ); // must be after SetFrameBufferDim() to have effect

	 void SendSettings( bool waitForEnd = false, bool flushCache = true );
	 void SendSettings( CSCDmaPacket& packet );
	 void SendSettings( CVifSCDmaPacket& packet );

	 // accessors
	 tU32 GetFrameBufferAddr( void ) { return gsrFrame.FBP * 2048; }

      protected:
	 tSourceChainTag SettingsDmaTag;
	 tGifTag	SettingsGifTag;
	 ps2_gsreg_frame	gsrFrame;	tU64 FrameAddr;
	 ps2_gsreg_zbuf		gsrZBuf;	tU64 ZBufAddr;
	 ps2_gsreg_xyoffset	gsrXYOffset;	tU64 XYOffsetAddr;
	 ps2_gsreg_prmodecont	gsrPrModeCont;	tU64 PrModeContAddr;
	 ps2_gsreg_colclamp	gsrColClamp;	tU64 ColClampAddr;
	 ps2_gsreg_test		gsrTest;	tU64 TestAddr;
	 ps2_gsreg_alpha	gsrAlpha;	tU64 AlphaAddr;
	 ps2_gsreg_pabe		gsrPABE;	tU64 PABEAddr;
	 ps2_gsreg_fba		gsrFBA;		tU64 FBAAddr;
	 ps2_gsreg_dthe		gsrDTHE;	tU64 DTHEAddr;
	 ps2_gsreg_dimx      	gsrDIMX;	tU64 DIMXAddr;
	 ps2_gsreg_scissor	gsrScissor;	tU64 ScissorAddr;
	 ps2_gsreg_fogcol	gsrFogCol;	tU64 FogColAddr;
	 tU32			uiNumGSRegs;

	 CSCDmaPacket	GifPacket;

	 tZTestPassMode eZTestPassMode;
	 tU32		uiFBWidth, uiFBHeight;
	 float		InterlacedOffset;

      private:
	 void InitCommon( GS::tContext context );

   } __attribute__ ((aligned(16)));

   /********************************************
    * inline methods
    */

   inline void
   CDrawEnv::SetAlphaBlendFunc( tAlphaBlendVal a, tAlphaBlendVal b, tAlphaBlendVal c, tAlphaBlendVal d, tU32 fix )
   {
      gsrAlpha.A = a;
      gsrAlpha.B = b;
      gsrAlpha.C = c;
      gsrAlpha.D = d;
      gsrAlpha.FIX = fix;
   }
   
   inline void 
   CDrawEnv::SetDitherMatrix( tDitherVal dm00, tDitherVal dm01, tDitherVal dm02, tDitherVal dm03, 
			      tDitherVal dm10, tDitherVal dm11, tDitherVal dm12, tDitherVal dm13, 
			      tDitherVal dm20, tDitherVal dm21, tDitherVal dm22, tDitherVal dm23, 
			      tDitherVal dm30, tDitherVal dm31, tDitherVal dm32, tDitherVal dm33 )
   {
      gsrDIMX.DM00 = (tU64)dm00; gsrDIMX.DM01 = (tU64)dm01; gsrDIMX.DM02 = (tU64)dm02; gsrDIMX.DM03 = (tU64)dm03;
      gsrDIMX.DM10 = (tU64)dm10; gsrDIMX.DM11 = (tU64)dm11; gsrDIMX.DM12 = (tU64)dm12; gsrDIMX.DM13 = (tU64)dm13;  
      gsrDIMX.DM20 = (tU64)dm20; gsrDIMX.DM21 = (tU64)dm21; gsrDIMX.DM22 = (tU64)dm22; gsrDIMX.DM23 = (tU64)dm23;  
      gsrDIMX.DM30 = (tU64)dm30; gsrDIMX.DM31 = (tU64)dm31; gsrDIMX.DM32 = (tU64)dm32; gsrDIMX.DM33 = (tU64)dm33;  
   }	   

   inline void 
   CDrawEnv::SetScissorArea( tU32 scX, tU32 scY, tU32 scWidth, tU32 scHeight )
   {
      gsrScissor.SCAX0 = scX; gsrScissor.SCAY0 = scY;
      gsrScissor.SCAX1 = scX + scWidth - 1; gsrScissor.SCAY1 = scY + scHeight - 1;
   }

} // namespace GS

#endif // ps2s_drawenv_h

